// telnet.cpp : Defines the entry point for the DLL application.
//

#include "StdAfx.h"

#ifdef WIN32
#include "rexp.h"
#endif

#include "telnet.h"

CTelnet::CTelnet(
	const char*		host, 
	int				messageSize,
	const char*		prompt, 
	int				port, 
	double			loginTimeOut, 
	double			waitTimeOut
)
{ 
	if (messageSize < MIN_MSG_SIZE)
		messageSize = MIN_MSG_SIZE;

	m_messageSize = messageSize;
	p_message = new char[m_messageSize];

	memset(p_message, 0, sizeof(m_messageSize));

	strcpy(m_host,host);
	strcpy(m_prompt,prompt);

	m_port		= port;

	double _double_int, _double_dbl;

	_double_dbl = modf(loginTimeOut, &_double_int);

	m_loginWaitTimeOut.tv_sec = (int)_double_int;
	m_loginWaitTimeOut.tv_usec = (int)(_double_dbl *1000);

	_double_dbl = modf(waitTimeOut, &_double_int);

	m_waitTimeOut.tv_sec = (int)_double_int;
	m_waitTimeOut.tv_usec = (int)(_double_dbl *1000);

	m_isConnect = false;
	m_isSock	= false;
	m_isWaitTimeOut = false;

	strcpy(m_prompt, prompt);
#ifdef WIN32
	if (re_comp((unsigned char *)(m_prompt)) != NULL)
#else
	if (re_comp((char *)m_prompt) != 0)
#endif
	{
		strcpy(p_message, "CTelnet::CTelnet()	re_comp() error.");
		return;
	}


    int err;

#ifdef WIN32

    WORD wVersionRequested;
    WSADATA wsaData;

 
    wVersionRequested = MAKEWORD( 2, 0 );
 
    err = WSAStartup( wVersionRequested, &wsaData );
 
	if ( err != 0 )
	{
		strcpy(p_message, "CTelnet::CTelnet()	WSAStartup() error.");
		return;
	}
 
	if (LOBYTE( wsaData.wVersion ) != 2 ||
		HIBYTE( wsaData.wVersion ) != 0 ) 
	{
		WSACleanup();
		strcpy(p_message, "CTelnet::CTelnet()	winsock version error.");
		return;
	}
#endif


#ifdef WIN32
    m_sock = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
#else
    m_sock = socket(AF_INET, SOCK_STREAM, 0);
#endif


#ifdef WIN32
    if (m_sock == INVALID_SOCKET)
#else
    if (m_sock < 0)      
#endif
	{
		strcpy(p_message, "CTelnet::CTelnet()	socket() error.");
		return;
	}

	m_isSock	= true;

    struct sockaddr_in hostAddr;
    hostAddr.sin_family = AF_INET;
    hostAddr.sin_port = htons(TELNET_PORT);

 	unsigned long inetAddr = inet_addr(m_host);

	if (INADDR_NONE == inetAddr) 
	{
		struct hostent *hp = gethostbyname(m_host);
		if (NULL == hp)
		{
			strcpy(p_message, "CTelnet::CTelnet()	gethostbyname() error.");
			return;
		}

			hostAddr.sin_addr.s_addr = 
			((struct in_addr *)hp->h_addr)->s_addr;
	}
	else 
		hostAddr.sin_addr.s_addr = inetAddr;

#ifdef WIN32
    if (m_sock == NULL || m_sock == SOCKET_ERROR)
#else
    if (m_sock < 0)
#endif
	{
		strcpy(p_message, "CTelnet::CTelnet() any socket error.");
		return;
	}

	err = ::connect(m_sock, (struct sockaddr *)&hostAddr, sizeof(hostAddr));

#ifdef WIN32
	if (err == SOCKET_ERROR)
#else
	if (err < 0)
#endif
	{
		strcpy(p_message, "CTelnet::CTelnet()	connect() error.");
		return;
	}

	m_isConnect = true;
	return; 
}


CTelnet::~CTelnet()
{
#ifdef WIN32
	WSACleanup();

	if (m_isSock)
	{
		closesocket(m_sock);
	}
	
#else
	if (m_isSock)
	{
		close(m_sock);
	}
#endif

	delete [] p_message;
}

bool
CTelnet::login(const char* user, const char* password)
{
	memset(p_message, 0, sizeof(m_messageSize));
	
	
	if (waitfor("ogin:", m_loginWaitTimeOut) == false)
	{
		sprintf(p_message, "Error : CTelnet::login()    ...during waiting 'login:'");
		return false;
	}

	char buf[COMMAND_LENGTH];
	memset(buf,0,sizeof(buf));

	strcpy(buf, user);
	strcat(buf, "\r\n");

	int err = ::send(m_sock, buf, strlen(buf), 0);
    if (err != (int)strlen(buf))
	{
		strcpy(p_message, "CTelnet::login() send error");
		return false;
	}

	if (strcmp(password,"") != 0)
	{
		if (waitfor("assword:", m_loginWaitTimeOut) == false)
		{
			sprintf(p_message, "CTelnet::login() error   ...during waiting 'password:'");
			return false;
		}

		if (cmd(password, m_loginWaitTimeOut) == false)
		{
			sprintf(p_message, "CTelnet::login() error  ...after input password");
			return false;
		}
	}

	return true;
}

bool
CTelnet::cmd(const char* str, timeval waitTimeOut)
{
	timeval _tmp_m_waitTimeOut = m_waitTimeOut;
	m_waitTimeOut = waitTimeOut;

	bool truefalse = cmd(str);

	m_waitTimeOut = _tmp_m_waitTimeOut;

	return truefalse;
}


bool
CTelnet::cmd(const char* str)
{
	bool _waitPromptFlag = false;

	m_isWaitTimeOut = false;

	memset(p_message, 0, sizeof(m_messageSize));

	char _msgchk[41];
	memset(_msgchk, 0, sizeof(_msgchk));


	char buf[COMMAND_LENGTH];
	memset(buf,0,sizeof(buf));

	strcpy(buf, str);
	strcat(buf, "\r\n");

	int err = ::send(m_sock, buf, strlen(buf), 0);
    if (err != (int)strlen(buf))
	{
		sprintf(p_message, "CTelnet::cmd() send failed");
		return false;
	}
	
	fd_set readMask, writeMask, expMask ;

	FD_ZERO(&readMask);
	FD_ZERO(&writeMask);
	FD_ZERO(&expMask);

	FD_SET(m_sock,&readMask); 
	

	unsigned char buff;
	int _counter = 0;
	
	while(1)
	{
		_counter++;
		if (_counter >= m_messageSize)
		{
			strncpy(p_message, "CTelnet: Message Buffer Flow happened.",
				m_messageSize -1 );
			return false;
		}
		
		if ((buff = getTelnetChar()) == (unsigned char)(-1)) 
		{
			sprintf(p_message, "CTelnet::cmd() Recv error happened.");
			return false;
		}
		else if (buff == (unsigned char)(-2)) 
		{
			if (_waitPromptFlag == true)
			{					
				return true;
			}
			else
			{
				m_isWaitTimeOut = true;
				sprintf(p_message, "CTelnet::cmd() Timeout error happened.");
				return false;
			}
		}
		else
		{
			_waitPromptFlag = false;
			strncat(p_message,(const char*)&buff,1);		
		}
		
		int index;
		
		if (_counter >MSGCHK_SIZE) 
			index = MSGCHK_SIZE;
		else 
			index = _counter;
		
		memset(_msgchk + MSGCHK_SIZE + 1, 0, 1);
		memcpy(_msgchk, p_message + (_counter - index), index);
		
		if (re_exec((unsigned char *)(_msgchk)) )
			_waitPromptFlag = true;
		
	} // while(1)
}

bool 
CTelnet::cmd(const char* str, const char* waitstr, timeval waitTimeOut)
{
	timeval _tmp_m_waitTimeOut = m_waitTimeOut;
	m_waitTimeOut = waitTimeOut;

	bool truefalse = cmd(str, waitstr);

	m_waitTimeOut = _tmp_m_waitTimeOut;

	return truefalse;
}

bool
CTelnet::cmd(const char* str, const char* waitstr)
{
	if (strlen(waitstr) == NULL)
	{
		sprintf(p_message, "CTelnet::cmd() String length is 0");
		return false;
	}

	bool _waitPromptFlag = false;

	m_isWaitTimeOut = false;

	memset(p_message, 0, sizeof(m_messageSize));

	char _msgchk[41];
	memset(_msgchk, 0, sizeof(_msgchk));


	char buf[COMMAND_LENGTH];
	memset(buf,0,sizeof(buf));

	strcpy(buf, str);
	strcat(buf, "\r\n");

	int err = ::send(m_sock, buf, strlen(buf), 0);
    if (err != (int)strlen(buf))
	{
		sprintf(p_message, "CTelnet::cmd() send failed");
		return false;
	}
	
	fd_set readMask, writeMask, expMask ;

	FD_ZERO(&readMask);
	FD_ZERO(&writeMask);
	FD_ZERO(&expMask);

	FD_SET(m_sock,&readMask); 
	

	unsigned char buff;
	int _counter = 0;
	
	while(1)
	{
		_counter++;
		if (_counter >= m_messageSize)
		{
			strncpy(p_message, "CTelnet: Message Buffer Flow happened.",
				m_messageSize -1 );
			return false;
		}
		
		if ((buff = getTelnetChar()) == (unsigned char)(-1)) 
		{
			sprintf(p_message, "CTelnet::cmd() Recv error happened.");
			return false;
		}
		else if (buff == (unsigned char)(-2)) 
		{
			if (_waitPromptFlag == true)
			{					
				return true;
			}
			else
			{
				m_isWaitTimeOut = true;
				sprintf(p_message, "CTelnet::cmd() Timeout error happened.");
				return false;
			}
		}
		else
		{
			_waitPromptFlag = false;
			strncat(p_message,(const char*)&buff,1);		
		}
		
		int index;
		
		if (_counter >MSGCHK_SIZE) 
			index = MSGCHK_SIZE;
		else 
			index = _counter;
		
		memset(_msgchk + MSGCHK_SIZE + 1, 0, 1);
		memcpy(_msgchk, p_message + (_counter - index), index);
		
		//if (re_exec((unsigned char *)(_msgchk)) )
		if (strstr(_msgchk, waitstr) != NULL)
			_waitPromptFlag = true;
		
	} // while(1)
}


bool 
CTelnet::waitfor(const char* str, timeval waitTimeOut)
{
	timeval _tmp_m_waitTimeOut = m_waitTimeOut;
	m_waitTimeOut = waitTimeOut;

	bool truefalse = waitfor(str);

	m_waitTimeOut = _tmp_m_waitTimeOut;

	return truefalse;
}



bool 
CTelnet::waitfor(const char* str)
{
	m_isWaitTimeOut = false;

	memset(p_message,0,sizeof(m_messageSize));

	int matchIdx = 0, maxIdx = strlen(str);

	fd_set readMask, writeMask, expMask ;

	FD_ZERO(&readMask);
	FD_ZERO(&writeMask);
	FD_ZERO(&expMask);

	FD_SET(m_sock,&readMask); 
	
	
	unsigned char buff;
	int _counter = 0;
	
	do
	{
		_counter++;
		if (_counter >= m_messageSize)
		{
			sprintf(p_message, "CTelnet: Message Buffer Flow happened.");
			return false;
		}
		
		if ((buff = getTelnetChar()) == (unsigned char)(-1)) 
		{
			sprintf(p_message, "CTelnet::waitfor  Recv error happened.");
			return false;
		}
		else if (buff == (unsigned char)(-2)) 
		{
			m_isWaitTimeOut = true;
			sprintf(p_message, "CTelnet::waitfor Timeout happened.");
			return false;
		}
		else
		{
			if (buff == str[matchIdx])
				matchIdx++;
			else
				matchIdx = 0;
		}
	} while (matchIdx != maxIdx);


	return true;
}



bool
CTelnet::isConnect()
{
	return m_isConnect;
}


unsigned char
CTelnet::getTelnetChar(void)
{
	bool iacFlag = false;
    bool skipFlag = false;
	bool replyFlag = false;

    char buf[2];
	buf[0] = buf[1] = 0;
	
	char wont[] = { TELNET_IAC, TELNET_WONT, 0 }; 


	fd_set readMask, writeMask, expMask ;

	FD_ZERO(&readMask);
	FD_ZERO(&writeMask);
	FD_ZERO(&expMask);

	FD_SET(m_sock,&readMask); 
	
	for (;;) 
	{
		timeval _waitTimeOut = m_waitTimeOut;

		if (select(m_sock+1,&readMask, &writeMask, &expMask, &_waitTimeOut)
			== 0) // timeout
		{
			return((unsigned char)(-2));
		}

#ifdef WIN32
		if (recv(m_sock, buf, 1, 0) == SOCKET_ERROR) 
#else
		if (recv(m_sock, buf, 1, 0) < 0) 
#endif
		{
			return((unsigned char)(-1));
		}

		if (skipFlag == true) 
		{
			skipFlag = false;
			continue;
		}
		else if	(replyFlag == true)
		{
			replyFlag = false;

			wont[2] = buf[0];
			::send(m_sock, wont, 3, 0);

			continue;
		
		} 
		else if (iacFlag == true) 
		{
			if ((unsigned char)buf[0] == (unsigned char)TELNET_DO) 
			{
				replyFlag = true;
			} 
			else if (((unsigned char)buf[0] == (unsigned char)TELNET_DONT) ||
				   ((unsigned char)buf[0] == (unsigned char)TELNET_WILL) ||
				   ((unsigned char)buf[0] == (unsigned char)TELNET_WONT)) 
			{
				skipFlag = true;
			}
			
			iacFlag = false;
		}
		else if ((unsigned char)buf[0] == (unsigned char)TELNET_IAC) 
		{
			iacFlag = true;
			continue;
		} 
		else 
			return (unsigned char)buf[0];
    }
}

char* 
CTelnet::getmsg(void)
{
	return p_message;
}

